@@ -5,18 +5,24 @@
 #    |_||___/ \__||_|
 # strings you can count on
 
-__version__ = "1.0.2"
+__version__ = "1.1.7"
 import functools
-import math
+import itertools
+import types
+import sys
+import inspect
 
 """
 Note: the changelog is now in changelog.md
 
-You can view the changelog on www.salabim/istr_changelog.html
+You can view the changelog on www.salabim.org/istr/changelog
+
+The readme can be viewed on www.salabim.org/istr/
 """
 
 _0_to_Z = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
 
+
 class _range:
     """
     based on https://codereview.stackexchange.com/questions/229073/pure-python-range-implementation
@@ -34,6 +40,12 @@ class _range:
             step_sign = 1
         self._len = max(1 + (self.stop - self.start - step_sign) // self.step, 0)
         self.parent_cls = cls
+        self.init_done = True
+
+    def __setattr__(self, name, value):
+        if getattr(self, "init_done", False):
+            raise AttributeError()
+        super().__setattr__(name, value)
 
     def __contains__(self, value):
         if isinstance(value, int):
@@ -169,27 +181,29 @@ class istr(str):
     ----------
     value : any
         if str the value will to be interpreted as an int
-            istr("8") ==> istr("8")
+            istr('8') ==> istr('8')
         if numeric, the value will be interpreted as an int
-            istr(8) ==> istr("8")
-        if a dict (or subtype of dict), the same type dict will be returned with all values istr"ed
-            istr({0: 0, 1: 1, 2: 4}) ==> {0: istr("0"), 1: istr("1"), 2: istr("4")}
+            istr(8) ==> istr('8')
+        if a dict (or subtype of dict), the same type dict will be returned with all values istr'ed
+            istr({0: 0, 1: 1, 2: 4}) ==> {0: istr('0'), 1: istr('1'), 2: istr('4')}
         if an iterator, the iterator will be mapped with istr
             istr(i * i for i in range(3)) ==> <map object>
-            list(istr(i * i for i in range(3))) ==> [istr("0"), istr("1"), istr("4")]
-        if an iterable, the same type will be returned with all elements istr"ed
-            istr([0, 1, 4]) ==> [istr("0"), istr("1"), istr("4")]
-            istr((0, 1, 4)) ==> (istr("0"), istr("1"), istr("4"))
-            istr({0, 1, 4}) ==> {istr("4"), istr("0"), istr("1")} # or similar
+            list(istr(i * i for i in range(3))) ==> [istr('0'), istr('1'), istr('4')]
+        if an iterable, the same type will be returned with all elements istr'ed
+            istr([0, 1, 4]) ==> [istr('0'), istr('1'), istr('4')]
+            istr((0, 1, 4)) ==> (istr('0'), istr('1'), istr('4'))
+            istr({0, 1, 4}) ==> {istr('4'), istr('0'), istr('1')} # or similar
         if a range, an istr.range instance will be returned
             istr(range(3)) ==> istr.range(3)
-            list(istr(range(3))) ==> [istr("0"), istr("1"), istr("2")]
+            list(istr(range(3))) ==> [istr('0'), istr('1'), istr('2')]
             len(istr(range(3))) ==> 3
+        if an istr, the same istr will be returned istr(istr('2')) ==> istr('2')
 
         it is possible to give more than one parameter, in which case a tuple
         of the istrs of the parameters will be returned, which can be handy
         to multiple assign, e.g.
-            a, b, c = istr(5, 6, 7) ==> a=istr("5") , b=istr("6"), c=istr("7")"""
+            a, b, c = istr(5, 6, 7) ==> a=istr('5') , b=istr('6'), c=istr('7')
+    """
 
     __slots__ = ("_as_int", "_as_repr")
 
@@ -197,7 +211,6 @@ class istr(str):
     _repr_mode = "istr"
     _base = 10
     _nan = object()
-    _force_istr_repr = False
     _digits_cache = {}
 
     @staticmethod
@@ -220,28 +233,6 @@ class istr(str):
         except:
             return cls._nan
 
-    def _check_is_int(self, *args, right=False):
-        operator = args[len(args) == 2]
-        if len(args) == 2:
-            other = args[0]
-            if not self.is_int() or self._to_int(other) is self._nan:
-                self.__class__._force_istr_repr = True
-                if right:
-                    message = f"unsupported operand for {operator}: {repr(other)} and {repr(self)}"
-                else:
-                    message = f"unsupported operand for {operator}: {repr(self)} and {repr(other)}"
-                self.__class__._force_istr_repr = False
-                raise TypeError(message)
-        else:
-            if not self.is_int():
-                self.__class__._force_istr_repr = True
-                message = f"unsupported operand for {operator}: {repr(self)}"
-                self.__class__._force_istr_repr = False
-                raise TypeError(message)
-
-    def _rcheck_is_int(self, *args):
-        self._check_is_int(*args, right=True)
-
     def __new__(cls, *value):
         if len(value) == 0:
             raise TypeError("no parameter given")
@@ -260,6 +251,8 @@ class istr(str):
                 return map(functools.partial(cls), value)
             return type(value)(map(functools.partial(cls), value))
 
+        if isinstance(value, str) and value.startswith('='):
+            value=str(cls.compose(value[1:],inspect.currentframe().f_back.f_back.f_globals))   
         as_int = cls._to_int(value)
         if isinstance(value, str):
             as_str = value
@@ -304,163 +297,193 @@ class istr(str):
     def __ne__(self, other):
         return not self == other
 
-    def __contains__(self, other):
-        return super().__contains__(str(other))
-
     def __repr__(self):
-        if self._force_istr_repr:
-            return f"{self.__class__.__name__}({super().__repr__()})"
         return self._as_repr
 
-    def __le__(self, other):
-        self._check_is_int(other, "<=")
-        return self._as_int <= self._to_int(other)
-
-    def __lt__(self, other):
-        self._check_is_int(other, "<")
-        return self._as_int < self._to_int(other)
-
-    def __ge__(self, other):
-        self._check_is_int(other, ">=")
-        return self._as_int >= self._to_int(other)
-
-    def __gt__(self, other):
-        self._check_is_int(other, ">")
-        return self._as_int > self._to_int(other)
-
     def __bool__(self):
         if self.is_int():
             return bool(self._as_int)
         return bool(str(self))
 
-    def __add__(self, other):
-        self._check_is_int(other, "+")
-        return self.__class__(self._as_int + self._to_int(other))
+    def _frepr(self, obj):
+        # like repr, but if obj is an istr, the as_repr is not used to make sure the
+        # the returned value is istr(...) and not infuenced by the repr mode
+        if isinstance(obj, self.__class__):
+            return f"{obj.__class__.__name__}({super(istr, obj).__repr__()})"
+        return repr(obj)
 
-    def __radd__(self, other):
-        self._rcheck_is_int(other, "+")
-        return self.__class__(self._to_int(other) + self._as_int)
+    def _int_method(self, name, op, *args):
+        if len(args) == 1:
+            other = args[0]
+            if not self.is_int() or self._to_int(other) is self._nan:
+                if name.startswith("__r"):
+                    raise TypeError(f"unsupported operand for {op}: {self._frepr(other)} and {self._frepr(self)}")
+                else:
+                    raise TypeError(f"unsupported operand for {op}: {self._frepr(self)} and {self._frepr(other)}")
+            if "<" in op or ">" in op:
+                return getattr(self._as_int, name)(self._to_int(other))
+            else:
+                return self.__class__(getattr(self._as_int, name)(self._to_int(other)))
+        else:
+            if not self.is_int():
+                raise TypeError(f"unsupported operand for {op}: {self._frepr(self)}")
+            return self.__class__(getattr(self._as_int, name)())
 
-    def __sub__(self, other):
-        self._check_is_int(other, "-")
-        return self.__class__(self._as_int - self._to_int(other))
+    for name_op in (
+        "__add__+ __radd__+ __sub__- __rsub__- __mul__* __rmul__* __floordiv__// __rfloordiv__// "
+        "__truediv__/ __rtruediv__/ __pow__** __rpow__** __mod__% __rmod__% "
+        "__divmod__divmod __rdivmod__divmod "
+        "__le__<= __lt__< __gt__> __ge__>= "
+        "__round__round __trunc__trunc __floor__floor __ceil__ceil __neg__- __pos__+ "
+        "__invert__~ __abs__abs "
+    ).split():
+        i = len(name_op) - "".join(reversed(name_op)).find("_")  # pos of last _
+        name = name_op[:i]
+        op = name_op[i:]
 
-    def __rsub__(self, other):
-        self._rcheck_is_int(other, "-")
-        return self.__class__(self._to_int(other) - self._as_int)
-
-    def __mul__(self, other):
-        self._check_is_int(other, "*")
-        return self.__class__(self._as_int * self._to_int(other))
-
-    def __rmul__(self, other):
-        self._rcheck_is_int(other, "*")
-        return self.__class__(self._to_int(other) * self._as_int)
-
-    def __floordiv__(self, other):
-        self._check_is_int(other, "//")
-        return self.__class__(self._as_int // self._to_int(other))
-
-    def __rfloordiv__(self, other):
-        self._rcheck_is_int(other, "//")
-        return self.__class__(self._to_int(other) // self._as_int)
-
-    def __truediv__(self, other):
-        self._check_is_int(other, "/")
-        return self.__class__(self._as_int // self._to_int(other))
-
-    def __rtruediv__(self, other):
-        self._rcheck_is_int(other, "/")
-        return self.__class__(self._to_int(other) // self._as_int)
-
-    def __pow__(self, other):
-        self._check_is_int(other, "**")
-        return self.__class__(self._as_int ** self._to_int(other))
-
-    def __rpow__(self, other):
-        self._rcheck_is_int(other, "**")
-        return self.__class__(self._to_int(other) ** self._as_int)
-
-    def __mod__(self, other):
-        self._check_is_int(other, "%")
-        return self.__class__(self._as_int % self._to_int(other))
-
-    def __rmod__(self, other):
-        self._rcheck_is_int(other, "%")
-        return self.__class__(self._to_int(other) % self._as_int)
-
-    def __or__(self, other):
-        return self.__class__("".join((self, self.__class__(other))))
-
-    def __ror__(self, other):
-        return self.__class__("".join((self.__class__(other), self)))
-
-    def __round__(self):
-        self._check_is_int("round")
-        return self.__class__(round(self._as_int))
+        locals()[name] = functools.partialmethod(_int_method, name, op)
 
     def __int__(self):
         if not self.is_int():
-            raise ValueError(f"invalid literal for int() with base 10: {repr(self)}")
+            raise ValueError(f"invalid literal for int() with base 10: {self._frepr(self)}")
         return int(self._as_int)
 
-    def __trunc__(self):
-        self._check_is_int("trunc")
-        return self.__class__(math.trunc(self._as_int))
-
-    def __floor__(self):
-        self._check_is_int("floor")
-        return self.__class__(math.floor(self._as_int))
-
-    def __ceil__(self):
-        self._check_is_int("ceil")
-        return self.__class__(math.ceil(self._as_int))
-
-    def __matmul__(self, other):
-        return self.__class__(super().__mul__(other))
-
-    def __rmatmul__(self, other):
-        return self.__class__(super().__rmul__(other))
-
-    def __divmod__(self, other):
-        self._check_is_int(other, "divmod")
-        return self.__class__(divmod(self._as_int, self._to_int(other)))
-
-    def __rdivmod__(self, other):
-        self._rcheck_is_int(other, "divmod")
-        return self.__class__(divmod(self._to_int(other), self._as_int))
-
-    def __neg__(self):
-        self._check_is_int("-")
-        return self.__class__(-self._as_int)
-
-    def __pos__(self):
-        self._check_is_int("+")
-        return self
-
-    def __abs__(self):
-        self._check_is_int("abs")
-        return self.__class__(abs(self._as_int))
-
     def is_even(self):
-        self._check_is_int("is_even")
-        return self._as_int % 2 == 0
+        return istr.is_divisible_by(self,2)
 
     def is_odd(self):
-        self._check_is_int("is_odd")
-        return self._as_int % 2 == 1
+        return not istr.is_divisible_by(self,2)
+
+    def is_divisible_by(self, divisor):
+        return istr.interpret_as_int(self) % int(divisor) == 0
+
+    def is_square(self):
+        return istr.is_power_of(self, 2)
+
+    def is_cube(self):
+        return istr.is_power_of(self, 3)
+
+    def is_power_of(self, exponent):
+        n = istr.interpret_as_int(self)
+        if exponent < 1:
+            raise ValueError(f"exponent must be >=1; not {exponent}")
+        if not isinstance(exponent, int):
+            raise TypeError(f"exponent must be int; not {type(exponent)}")
+        return n >= 0 and self == round(n ** (1 / exponent)) ** exponent
+
+    def is_prime(self):
+        n = istr.interpret_as_int(self)
+        if n < 2:
+            return False
+        if n == 2:
+            return True
+        if not n & 1:
+            return False
+
+        for x in range(3, int(n**0.5) + 1, 2):
+            if n % x == 0:
+                return False
+        return True
+
+    def decompose(self, letters, namespace=None):
+        """
+        decompose one-letter variables into global variables
+        each one-letter variable must represent just one character
+        same one-letter variables represent the the same character
+        the istr must have the same length as the letters
+        """
+        if namespace is None:
+            namespace = inspect.currentframe().f_back.f_globals
+
+        lookup = {}
+
+        for letter, ch in zip(letters, self):
+            if letter in lookup and lookup[letter] != ch:
+                raise ValueError(f"multiple values found for variable {letter}")
+            if not letter.isidentifier():
+                raise ValueError(f"{letter} cannot be used as a variable")
+            lookup[letter] = ch
+        if len(letters) != len(self):
+            raise ValueError(f"incorrect number of variables {len(letters)}; should be {len(self)}")
+        namespace.update(lookup)
+
+    @classmethod
+    def compose(cls, letters, namespace=None):
+        """
+        compose an istr from individual letter variables
+        """
+        if namespace is None:
+            namespace = inspect.currentframe().f_back.f_globals
+        for letter in letters:
+            if letter not in namespace:
+                raise ValueError(f"variable {letter} not defined")
+
+        return istr("").join(istr(namespace[letter]) for letter in letters)
+
+    def __or__(self, other):
+        try:
+            return self.__class__(str(self).__add__(other))
+        except TypeError:
+            raise TypeError(f"unsupported operand type(s) for |: {self._frepr(self)} and {self._frepr(other)}")
+
+    def __ror__(self, other):
+        try:
+            return self.__class__(other.__add__(str(self)))
+        except TypeError:
+            raise TypeError(f"unsupported operand type(s) for |: {self._frepr(other)} and {self._frepr(self)}")
+
+    def __matmul__(self, other):
+        try:
+            return self.__class__(super().__mul__(other))
+        except TypeError:
+            raise TypeError(f"unsupported operand type(s) for @: {self._frepr(self)}  and {self._frepr(other)}")
+
+    def __rmatmul__(self, other):
+        try:
+            return self.__class__(super().__rmul__(other))
+        except TypeError:
+            raise TypeError(f"unsupported operand type(s) for @|: {self._frepr(other)}  and {self._frepr(self)}")
+
+    def __getitem__(self, key):
+        return self.__class__(super().__getitem__(key))
 
     def all_distinct(self):
         return len(self) == len(set(self))
 
-    def is_int(self):
-        return self._as_int is not self._nan
-
     def reversed(self):
         return self[::-1]
 
-    def __getitem__(self, key):
-        return self.__class__(super().__getitem__(key))
+    def interpret_as_int(self):
+        if isinstance(self, istr):
+            if not self.is_int():
+                raise TypeError(f"not interpretable as int: {self._frepr(self)}")
+            n = self._as_int
+        else:
+            n = int(self)
+        return n
+
+    def _str_method(self, name, *args, **kwargs):
+        return self.__class__(getattr(super(), name)(*args, **kwargs))
+
+    for name in (
+        "capitalize casefold center expandtabs format join ljust lower lstrip partition removeprefix "
+        "removesuffix replace rjust rpartition rsplit rstrip split strip swapcase title translate upper zfill"
+    ).split():
+        locals()[name] = functools.partialmethod(_str_method, name)
+
+    @classmethod
+    def _itertools_method(cls, name, *args, **kwargs):
+        return cls(getattr(itertools, name)(*args, **kwargs))
+
+    for name in dir(itertools):
+        if not name.startswith("__"):
+            if name in ("groupby", "tee"):
+                locals()[name] = getattr(itertools, name)
+            else:
+                locals()[name] = functools.partialmethod(_itertools_method, name)
+
+    def is_int(self):
+        return self._as_int is not self._nan
 
     @classmethod
     def concat(cls, iterable):
@@ -486,8 +509,7 @@ class istr(str):
 
             cls._int_format = int_format
 
-        def __enter__(self):
-            ...
+        def __enter__(self): ...
 
         def __exit__(self, exc_type, exc_value, exc_tb):
             self.saved_cls._int_format = self.saved_int_format
@@ -497,6 +519,8 @@ class istr(str):
         def __new__(cls, cls_repr_mode, mode=None):
             if mode is None:
                 return cls_repr_mode._repr_mode
+            if mode == int:
+                mode = "int"
             if mode in ("istr", "str", "int"):  # _istr is used only for TypeErrors
                 return super().__new__(cls)
             raise TypeError(f"mode not 'istr', 'str' or 'int', but {repr(mode)}")
@@ -506,8 +530,7 @@ class istr(str):
             self.saved_cls = cls
             cls._repr_mode = mode
 
-        def __enter__(self):
-            ...
+        def __enter__(self): ...
 
         def __exit__(self, exc_type, exc_value, exc_tb):
             self.saved_cls._repr_mode = self.saved_repr_mode
@@ -526,8 +549,7 @@ class istr(str):
             self.saved_cls = cls
             cls._base = base
 
-        def __enter__(self):
-            ...
+        def __enter__(self): ...
 
         def __exit__(self, exc_type, exc_value, exc_tb):
             self.saved_cls._base = self.saved_base
@@ -544,7 +566,7 @@ class istr(str):
         if no args, 0-9 will be used
 
         all given args will be used
-        each arg has to be either null string, <digit>, <digit>-<digit> or -<digit>
+        each arg has to be either null string, <digit>, <digit>-<digit>, <digit>- or -<digit>
 
         the digits may be '0' through '9' and 'A' through 'Z' (not case sensitive)
         The returned value will always be in uppercase (if applicable).
@@ -569,7 +591,7 @@ class istr(str):
         """
         key = (args, cls._base, cls._int_format, cls._repr_mode)
         if key in cls._digits_cache:
-            return cls.digits_cache[key]
+            return cls._digits_cache[key]
         result = []
         if not args:
             args = ["0-9"]
@@ -605,89 +627,20 @@ class istr(str):
         cls._digits_cache[key] = result
         return result
 
-    def capitalize(self, *args, **kwargs):
-        return self.__class__(super().capitalize(*args, **kwargs))
 
-    def casefold(self, *args, **kwargs):
-        return self.__class__(super().casefold(*args, **kwargs))
-
-    def center(self, *args, **kwargs):
-        return self.__class__(super().center(*args, **kwargs))
-
-    def expandtabs(self, *args, **kwargs):
-        return self.__class__(super().expandtabs(*args, **kwargs))
-
-    def format(self, *args, **kwargs):
-        return self.__class__(super().format(*args, **kwargs))
-
-    def join(self, *args, **kwargs):
-        return self.__class__(super().join(*args, **kwargs))
-
-    def ljust(self, *args, **kwargs):
-        return self.__class__(super().ljust(*args, **kwargs))
-
-    def lower(self, *args, **kwargs):
-        return self.__class__(super().lower(*args, **kwargs))
-
-    def lstrip(self, *args, **kwargs):
-        return self.__class__(super().lstrip(*args, **kwargs))
-
-    def partition(self, *args, **kwargs):
-        return self.__class__(super().partition(*args, **kwargs))
-
-    def removeprefix(self, *args, **kwargs):
-        return self.__class__(super().removeprefix(*args, **kwargs))
-
-    def removesuffix(self, *args, **kwargs):
-        return self.__class__(super().removesuffix(*args, **kwargs))
-
-    def replace(self, *args, **kwargs):
-        return self.__class__(super().replace(*args, **kwargs))
-
-    def rjust(self, *args, **kwargs):
-        return self.__class__(super().rjust(*args, **kwargs))
-
-    def rpartition(self, *args, **kwargs):
-        return self.__class__(super().rpartition(*args, **kwargs))
-
-    def rsplit(self, *args, **kwargs):
-        return self.__class__(super().rsplit(*args, **kwargs))
-
-    def rstrip(self, *args, **kwargs):
-        return self.__class__(super().rstrip(*args, **kwargs))
-
-    def split(self, *args, **kwargs):
-        return self.__class__(super().split(*args, **kwargs))
-
-    def strip(self, *args, **kwargs):
-        return self.__class__(super().strip(*args, **kwargs))
-
-    def swapcase(self, *args, **kwargs):
-        return self.__class__(super().swapcase(*args, **kwargs))
-
-    def title(self, *args, **kwargs):
-        return self.__class__(super().title(*args, **kwargs))
-
-    def translate(self, *args, **kwargs):
-        return self.__class__(super().translate(*args, **kwargs))
-
-    def upper(self, *args, **kwargs):
-        return self.__class__(super().upper(*args, **kwargs))
-
-    def zfill(self, *args, **kwargs):
-        return self.__class__(super().zfill(*args, **kwargs))
+istr.type = type(istr(0))
 
 
-def main():
-    with istr.repr_mode("int"):
-        print(repr(istr(3)))
-        print(repr(istr("a")))
-        with istr.int_format("4"):
-            print(repr(istr(3)))
-            print(repr(istr("a")))
-        with istr.int_format("04"):
-            print(repr(istr(3)))
-            print(repr(istr("a")))
+class istrModule(types.ModuleType):
+    def __call__(self, *args, **kwargs):
+        return istr(*args, **kwargs)
 
-if __name__ == "__main__":
-    main()
+    def __setattr__(self, item, value):
+        setattr(istr, item, value)
+
+    def __getattr__(self, item):
+        return getattr(istr, item)
+
+
+if __name__ != "__main__":
+    sys.modules["istr"].__class__ = istrModule
