@@ -6,14 +6,13 @@ import re
 from pathlib import Path
 
 if __name__ == "__main__":  # to make the tests run without the pytest cli
-    file_folder = Path(__file__).parent
-    top_folder = (file_folder / "..").resolve()
-    sys.path.insert(0, str(top_folder))
+    file_folder = os.path.dirname(__file__)
     os.chdir(file_folder)
+    sys.path.insert(0, file_folder + "/../istr")
 
 import pytest
 
-from istr import istr
+import istr
 
 istr.equals = lambda self, other: type(self) == type(other) and (str(self) == str(other))
 # this method tests whether self and other are exactly the same
@@ -68,6 +67,7 @@ def test_arithmetic():
 
 
 def test_lt():
+    assert type(two < three) == bool
     assert two < three
     assert 2 < three
     assert two < 3
@@ -82,6 +82,7 @@ def test_lt():
 
 
 def test_le():
+    assert type(two <= three) == bool
     assert two <= three
     assert 2 <= three
     assert two <= 3
@@ -92,6 +93,7 @@ def test_le():
 
 
 def test_gt():
+    assert type(three > two) == bool
     assert three > two
     assert 3 > two
     assert three > 2
@@ -106,6 +108,7 @@ def test_gt():
 
 
 def test_ge():
+    assert type(three >= two) == bool
     assert three >= two
     assert 3 >= two
     assert three >= 2
@@ -188,7 +191,7 @@ def test_divmod():
     assert divmod(eleven, three) == (istr(3), istr(2))
     assert divmod(11, three) == (istr(3), istr(2))
     assert divmod(eleven, 3) == (istr(3), istr(2))
-    assert divmod(11, 3) == (3, 2)
+    assert divmod(11, 3) == (3, 2)  # just for reference
 
 
 def test_iter():
@@ -286,9 +289,6 @@ def test_int_format():
     assert istr(" 8 ") == " 8 "
 
 
-#    assert repr(istr('a')) == "istr('a')"
-
-
 def test_range_int_format():
     r = istr.range(11)
     assert repr(r) == "istr.range(0, 11)"
@@ -303,7 +303,96 @@ def test_even_odd():
     assert not istr(1).is_even()
 
     assert istr(12345678).is_even()
-    assert not istr(12345678).is_odd()
+    with pytest.raises(TypeError, match=re.escape(f"not interpretable as int")):
+        istr("a").is_odd()
+    with pytest.raises(TypeError, match=re.escape(f"not interpretable as int")):
+        istr("a").is_even()
+
+    assert istr.is_odd(1)
+    assert not istr.is_even(1)
+    assert istr.is_even(12345678)
+    assert istr.is_odd(11111111)
+
+def test_is_divisible():
+    assert istr(18).is_divisible_by(3)
+    assert istr(18).is_divisible_by(istr(3))
+    assert not istr(19).is_divisible_by(3)
+    assert not istr(19).is_divisible_by(istr(3))
+    with pytest.raises(TypeError, match=re.escape(f"not interpretable as int")):
+        istr("a").is_divisible_by(3)
+    assert istr.is_divisible_by(18, 3)
+    assert not istr.is_divisible_by(19, 3)
+
+def test_is_square():
+    assert not istr(-1).is_square()
+    assert istr(0).is_square()
+    assert istr(1).is_square()
+    assert not istr(2).is_square()
+    assert istr(4).is_square()
+    assert istr(16).is_square()
+    assert not istr(99).is_square()
+    with pytest.raises(TypeError, match=re.escape(f"not interpretable as int")):
+        istr("a").is_square()
+    assert istr.is_square(0)
+    assert istr.is_square(1)
+    assert not istr.is_square(2)
+    assert istr.is_square(4)
+    assert istr.is_square(16)
+
+def test_is_cube():
+    assert not istr(-1).is_cube()
+    assert istr(0).is_cube()
+    assert istr(1).is_cube()
+    assert not istr(2).is_cube()
+    assert istr(8).is_cube()
+    assert istr(27).is_cube()
+    assert not istr(99).is_cube()
+    with pytest.raises(TypeError, match=re.escape(f"not interpretable as int")):
+        istr("a").is_cube()
+    assert istr.is_cube(0)
+    assert istr.is_cube(1)
+    assert not istr.is_cube(2)
+    assert istr.is_cube(8)
+    assert istr.is_cube(27)
+
+
+def test_is_power_of():
+    assert not istr(-1).is_power_of(3)
+    assert istr(0).is_power_of(3)
+    assert istr(1).is_power_of(3)
+    assert not istr(2).is_power_of(3)
+    assert istr(8).is_power_of(3)
+    assert istr(27).is_power_of(3)
+    assert not istr(99).is_power_of(3)
+    with pytest.raises(TypeError, match=re.escape(f"not interpretable as int")):
+        istr("a").is_power_of(3)
+    assert istr.is_power_of(0,3)
+    assert istr.is_power_of(1,3)
+    assert not istr.is_power_of(2,3)
+    assert istr.is_power_of(8,3)
+    assert istr.is_power_of(27,3)
+    with pytest.raises(TypeError):
+        istr(1).is_power_of(3.1)
+    with pytest.raises(ValueError):
+        istr(1).is_power_of(-1)
+ 
+def test_is_prime():
+    assert not istr(0).is_prime()
+    assert not istr(1).is_prime()
+    assert istr(2).is_prime()
+    assert istr(3).is_prime()
+    assert not istr(4).is_prime()
+    assert istr(97).is_prime()
+    assert not istr(99).is_prime()
+    with pytest.raises(TypeError, match=re.escape(f"not interpretable as int")):
+        istr("a").is_prime()
+    assert not istr.is_prime(0)
+    assert not istr.is_prime(1)
+    assert istr.is_prime(2)
+    assert istr.is_prime(3)
+    assert not istr.is_prime(4)
+    assert istr.is_prime(97)
+    assert not istr.is_prime(99)
 
 
 def test_join():
@@ -314,17 +403,29 @@ def test_join():
     s = istr("").join(("4", "5", "6"))
     assert s == "456"
     assert s == 456
-    assert type(s) == istr
+    assert type(s) == istr.type
 
     s = istr("").join(istr(("4", "5", "6")))
     assert s == "456"
     assert s == 456
-    assert type(s) == istr
+    assert type(s) == istr.type
 
     s = istr("").join(istr(("", "", "6")))
     assert s == "6"
     assert s == 6
-    assert type(s) == istr
+    assert type(s) == istr.type
+
+
+def test_or():
+    assert (eleven | twelve).equals(istr("1112"))
+    assert ("11" | twelve).equals(istr("1112"))
+    assert (eleven | "12").equals(istr("1112"))
+    with pytest.raises(TypeError):
+        11 | twelve
+    with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for |: 11 and istr('12')")):
+        11 | twelve
+    with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for |: istr('11') and 12")):
+        eleven | 12
 
 
 def test_matmul():
@@ -369,9 +470,9 @@ def test_is_int():
         +istr("a")
     with pytest.raises(TypeError, match=re.escape(f"unsupported operand for -: istr('a')")):
         -istr("a")
-    with pytest.raises(TypeError, match=re.escape(f"unsupported operand for is_odd: istr('a')")):
+    with pytest.raises(TypeError):
         istr("a").is_odd()
-    with pytest.raises(TypeError, match=re.escape(f"unsupported operand for is_even: istr('a')")):
+    with pytest.raises(TypeError):
         istr("a").is_even()
     assert istr(1).is_int()
     assert not istr("a").is_int()
@@ -388,6 +489,9 @@ def test_is_int():
 
 def test_str():
     assert repr(str(five)) == "'5'"
+    assert f"|{istr(1234):6}|" == "|1234  |"
+    with pytest.raises(ValueError):
+        f"{istr(1234):d}"
 
 
 def test_bool():
@@ -406,16 +510,16 @@ def test_trunc_and_friends():
 
 
 def test_data_structures():
-    assert repr(istr(list(range(1, 4)))) == "[istr('1'), istr('2'), istr('3')]"
-    assert repr(istr(tuple(range(1, 4)))) == "(istr('1'), istr('2'), istr('3'))"
+    assert istr(list(range(1, 4))) == [istr("1"), istr("2"), istr("3")]
+    assert istr(tuple(range(1, 4))) == (istr("1"), istr("2"), istr("3"))
     assert istr(set(range(1, 4))) == {istr(1), istr(2), istr(3)}
 
-    assert repr(list(istr(range(1, 4)))) == "[istr('1'), istr('2'), istr('3')]"
+    assert list(istr(range(1, 4))) == [istr("1"), istr("2"), istr("3")]
 
-    assert repr(list(istr.enumerate("abc"))) == "[(istr('0'), 'a'), (istr('1'), 'b'), (istr('2'), 'c')]"
-    assert repr(list(istr.enumerate("abc", 1))) == "[(istr('1'), 'a'), (istr('2'), 'b'), (istr('3'), 'c')]"
+    assert list(istr.enumerate("abc")) == [(istr("0"), "a"), (istr("1"), "b"), (istr("2"), "c")]
+    assert list(istr.enumerate("abc", 1)) == [(istr("1"), "a"), (istr("2"), "b"), (istr("3"), "c")]
 
-    assert repr(istr(dict(zero=0, one=1, two=4))) == "{'zero': istr('0'), 'one': istr('1'), 'two': istr('4')}"
+    assert istr(dict(zero=0, one=1, two=4)) == {"zero": istr("0"), "one": istr("1"), "two": istr("4")}
 
 
 def test_indexing():
@@ -428,9 +532,9 @@ def test_indexing():
 
 def test_reverse():
     a = istr(12345)
-    assert a.reversed(), same(istr(54321))
-    a = istr(-120)
-    assert a.reversed(), same(istr("-210"))
+    assert a.reversed().equals(istr(54321))
+    a = istr(-123)
+    assert a.reversed().equals(istr("321-"))
 
 
 def test_edge_cases():
@@ -449,7 +553,12 @@ def test_unpacking():
     assert x.equals(istr(1))
     assert y.equals(istr(2))
     assert z.equals(istr(3))
+    del x,y,z
 
+    x, y, z = a
+    assert x.equals(istr(1))
+    assert y.equals(istr(2))
+    assert z=="3"
 
 def test_repr_mode():
     hundred = istr(100)
@@ -470,8 +579,8 @@ def test_repr_mode():
     assert repr(hundred) == "istr('100')"
 
     with istr.repr_mode("int"):
-        a = istr("a")
-    assert repr(a) == "nan"
+        a = istr("abc")
+    assert repr(a) == "?"
 
     assert istr.repr_mode() == "istr"
 
@@ -540,6 +649,9 @@ def test_base():
         assert a * a == 225
 
 
+
+
+
 def test_digits():
     assert istr.digits().equals(istr("0123456789"))
     assert istr.digits("").equals(istr("0123456789"))
@@ -548,34 +660,111 @@ def test_digits():
     assert istr.digits("-3").equals(istr("0123"))
     assert istr.digits("1-4", "6", "8-9").equals(istr("1234689"))
     assert istr.digits("1", "1-2", "1-3").equals(istr("112123"))
-    a= istr.digits("a")
+    a = istr.digits("a")
     assert a.equals(istr("A"))
     assert not a.is_int()
-    with istr.base(36): 
-        a=istr.digits("a")
+    with istr.base(36):
+        a = istr.digits("a")
         assert a == 10
-        assert a =="A"
+        assert a == "A"
         assert istr.digits("-a").equals(istr("0123456789A"))
         assert istr.digits("x-").equals(istr("XYZ"))
         assert istr.digits("B-d").equals(istr("BCD"))
         assert istr.digits("-").equals(istr("0123456789"))
     with istr.base(16):
         ef = istr.digits("e-f")
-        assert (ef+1).equals(istr("F0"))
+        assert (ef + 1).equals(istr("F0"))
         assert ef == 239
 
+
+def test_digits_cache():
+    d = istr.digits()
+    assert id(d) == id(istr.digits())
+    assert int(d) == 123456789
+
+    with istr.base(16):
+        d = istr.digits()
+        assert id(d) == id(istr.digits())
+    assert int(d) == 4886718345
+
+
+def test_itertools():
+    def list100(it):
+        # just iterates over the first 100 elements in the iterable
+        return [next(it) for _ in range(100)]
+
+    assert list(istr.accumulate((1, 3, 4))) == [istr("1"), istr("4"), istr("8")]
+    assert list(istr.chain(range(2), range(2, 5))) == [istr("0"), istr("1"), istr("2"), istr("3"), istr("4")]
+    assert list(istr.combinations(range(5), r=3)) == list(istr(itertools.combinations(range(5), r=3)))
+    assert list(istr.combinations_with_replacement(range(5), r=3)) == list(istr(itertools.combinations_with_replacement(range(5), r=3)))
+    assert list(istr.compress("123456", [1, 0, 1, 0, 1, 1])) == [istr("1"), istr("3"), istr("5"), istr("6")]
+    assert list100(istr.count()) == list100(istr(itertools.count()))
+    assert list100(istr.cycle(range(10))) == list100(istr(itertools.cycle(range(10))))
+    assert list(istr.dropwhile(lambda x: x < 5, [1, 4, 6, 4, 1])) == [istr("6"), istr("4"), istr("1")]
+    assert list(istr.filterfalse(lambda x: x % 2, range(10))) == [istr("0"), istr("2"), istr("4"), istr("6"), istr("8")]
+    assert list(istr.islice("123456", 2)) == [istr("1"), istr("2")]
+    assert list(istr.pairwise("1234")) == [(istr("1"), istr("2")), (istr("2"), istr("3")), (istr("3"), istr("4"))]
+    assert list(istr.permutations(range(5), 3)) == list(istr(itertools.permutations(range(5), 3)))
+    assert list(istr.product(range(5), range(4))) == list(istr(itertools.product(range(5), range(4))))
+    assert list100(istr.repeat(10)) == list100(istr(itertools.repeat(10)))
+    assert list(istr.starmap(pow, [(2, 5), (3, 2), (10, 3)])) == [istr("32"), istr("9"), istr("1000")]
+    assert list(istr.takewhile(lambda x: x < 5, [1, 4, 6, 3, 8])) == [istr("1"), istr("4")]
+    assert list(istr.zip_longest("123", "56", fillvalue="0")) == [(istr("1"), istr("5")), (istr("2"), istr("6")), (istr("3"), istr("0"))]
+
+
 def test_all_distinct():
     assert istr("abcdef").all_distinct()
     assert not istr("aabcdef").all_distinct()
     assert istr("").all_distinct()
 
+
 def test_subclassing():
-    class jstr(istr):
-        ...
+    class jstr(istr.type): ...
 
     assert jstr(5).equals(jstr(5))
     assert repr(jstr(*range(3))) == "(jstr('0'), jstr('1'), jstr('2'))"
 
 
+def test_decompose():
+    istr("123").decompose("xyz")
+    assert x == 1
+    assert y == 2
+    assert z == 3
+    istr(456).decompose("xyz")
+    assert x == 4
+    assert y == 5
+    assert z == 6
+    istr(1231).decompose("xyzx")
+    assert x == 1
+    assert y == 2
+    assert z == 3
+    namespace = {}
+    istr(123).decompose("xyz", namespace=namespace)
+    assert namespace == dict(x=istr(1), y=istr(2), z=istr(3))
+
+    with pytest.raises(ValueError):
+        istr(1234).decompose("xyzx")
+    with pytest.raises(ValueError):
+        istr(1234).decompose("xyz")
+    with pytest.raises(ValueError):
+        istr(12).decompose("xyz")
+    with pytest.raises(ValueError):
+        istr(123).decompose("xy1")
+        
+
+
+
+def test_compose():
+    x = 1
+    y = 2
+    z = 3
+    s = istr.compose("xyz")
+    assert s == 123
+    with pytest.raises(ValueError):
+        s = istr.compose("wxyz", globals())  # w is not defined
+    s = istr.compose("xyz", namespace=dict(x=3, y=istr(4), z="5"))
+    assert s == 345
+
+
 if __name__ == "__main__":
     pytest.main(["-vv", "-s", "-x", __file__])
